#include "server.h"
#include "io.h"
#include "callfunc.h"

#include <assert.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>

#define BUFFER_SIZE	(16*1024*1024)

#define assert_call(expression)                 \
do {                                            \
    int rc = (expression);                      \
    assert(rc == 0);                            \
} while(0)

typedef enum {
    FRAME_TYPE_ACCEPT=0,
    FRAME_TYPE_REJECT,
    FRAME_TYPE_CALL_RESULT,
    FRAME_TYPE_ENTERING_LOOP,
    FRAME_TYPE_EVENT,
} frame_type_t;

static int write_buffer(int out_fd, const void *buffer, unsigned long buffer_size)
{
    assert_call(cinvoke_writen(out_fd, PS(buffer_size)));
    assert_call(cinvoke_writen(out_fd, buffer, buffer_size));
    return 0;
}

static void event_queue_init(cinvoke_event_queue_t *event_queue)
{
    /* Must guarantee that at every given point in time, *event_queue->last_ptr
       is NULL, even if a signal occurs between these two statements */
    event_queue->first = NULL;
    event_queue->last_ptr = &event_queue->first;
}

int cinvoke_server_init(cinvoke_server_t *server, int in_fd, int out_fd)
{
    cinvoke_server_migrate_fds(server, in_fd, out_fd);
    server->call_buffer_offset = 0;
    server->call_buffer = malloc(BUFFER_SIZE);
    assert(NULL != server->call_buffer);

    /* We don't need to start using the out pipe here because he's not allowed
     * to use the cinvoke object at all yet... */
    assert_call(cinvoke_writen(server->out_fd, PS(server->call_buffer)));
    
    server->exit_status.should_exit = CINVOKE_FALSE;
    server->nest_level = 0;
    server->out_fd_ref_count = 0;

    event_queue_init(&server->event_queue);

    return 0;
}

void cinvoke_server_migrate_fds(cinvoke_server_t *server, int in_fd, int out_fd)
{
    server->in_fd = in_fd;
    server->out_fd = out_fd;
}

void cinvoke_server_fini(cinvoke_server_t *server)
{
    free(server->call_buffer);
    /* For diagnostic purposes: */
    server->call_buffer_offset = -1;
}

/* The out pipe must be marked "in use" here */
static void send_event(cinvoke_server_t *server, const void *buffer, unsigned long buffer_size)
{
    frame_type_t frame_type = FRAME_TYPE_EVENT;
    assert_call(cinvoke_writen(server->out_fd, PS(frame_type)));
    assert_call(write_buffer(server->out_fd, buffer, buffer_size));
}

static void server_empty_event_queue(cinvoke_server_t *server)
{
    struct cinvoke_event_list_node *current;
    struct cinvoke_event_list_node *next;

    /* Reset the queue and remember the first element before walking the list,
     * to ensure that if a signal occurs, and it tries to add to the queue, it
     * adds to a new queue, and not to the same queue we are walking and
     * freeing. */
    current = server->event_queue.first;
    event_queue_init(&server->event_queue);
    
    for(; NULL != current; current = next) {
        send_event(server, current->buffer, current->buffer_size);
        next = current->next;
        free(current);
    }
}

static void event_queue_add_event(cinvoke_server_t *server,
                                  const void *buffer, unsigned long buffer_size)
{
    cinvoke_event_queue_t *event_queue = &server->event_queue;
    struct cinvoke_event_list_node *new_node;
    struct cinvoke_event_list_node **last_ptr;
    
    new_node = malloc(CINVOKE_EVENT_LIST_NODE_SIZE(buffer_size));
    new_node->next = NULL;
    new_node->buffer_size = buffer_size;
    memcpy(new_node->buffer, buffer, buffer_size);

    last_ptr = event_queue->last_ptr;
    
    /* Any reentrant add up to this assignment will add after our last_ptr. */
    event_queue->last_ptr = &new_node->next;
    /* Any reentrant add from this assignment will add after our new_node
     * (which is fine and does not interest us). */
    
    /* Since everybody is now adding after the new_node, the first part of the
     * list, in which last_ptr is, is no longer being "touched" so we can
     * happily walk over it until we reach its actual end. */
    
    while(NULL != (*last_ptr)) {
        last_ptr = &((*last_ptr)->next);
    }
    
    /* Now link the two parts of the list */
    (*last_ptr) = new_node;
}

static void start_using_out_pipe(cinvoke_server_t *server)
{
    assert(0 == server->out_fd_ref_count);
    server->out_fd_ref_count += 1;
}

static void stop_using_out_pipe(cinvoke_server_t *server)
{
    assert(1 == server->out_fd_ref_count);
    for(;;) {
        server_empty_event_queue(server);
        server->out_fd_ref_count -= 1;
        if(NULL == server->event_queue.first) {
            /* There's nothing in the queue, and the refcount is now 0, so
             * nothing will be added to the queue anymore... */
            break;
        }
        /* Something was added to the queue between our emptying it, and the
         * refcount decrease, so retry emptying it */
        server->out_fd_ref_count += 1;
    }
}

int cinvoke_server_send_event(cinvoke_server_t *server, const void *buffer, unsigned long buffer_size)
{
    /* This code is not atomic, and not thread-safe. However, it is signal
     * safe, and that is its purpose.
     */

    server->out_fd_ref_count += 1;
    if(1 == server->out_fd_ref_count) {
        /* The ref count was 0, so we can write directly to the pipe, and
           another signal during this function may have queued more stuff, so
           we need to empty the queue afterwards. */
        send_event(server, buffer, buffer_size);
        stop_using_out_pipe(server);
    } else {
        /* The ref count is more than 1, so it was already in use, must queue,
         * and must not empty the queue */
        event_queue_add_event(server, buffer, buffer_size);
        server->out_fd_ref_count -= 1;
    }
    return 0;
}

int cinvoke_server_handle_request(cinvoke_server_t *server)
{
    unsigned long request_nest_level;
    request_type_t request_type;
    cinvoke_bool_t request_accepted;
    frame_type_t frame_type;
    
    assert_call(cinvoke_readn(server->in_fd, PS(request_nest_level)));
    assert(request_nest_level <= server->nest_level);

    request_accepted = (request_nest_level == server->nest_level);
    if(request_accepted) frame_type = FRAME_TYPE_ACCEPT;
    else                 frame_type = FRAME_TYPE_REJECT;

    start_using_out_pipe(server);
    assert_call(cinvoke_writen(server->out_fd, PS(frame_type)));
    stop_using_out_pipe(server);
    
    assert_call(cinvoke_readn(server->in_fd, PS(request_type)));
    assert(request_type < REQUEST_TYPE_COUNT);
    assert_call(cinvoke_server_request_handlers[request_type](server, request_type, request_accepted));

    return 0;
}

static int server_loop(cinvoke_server_t *server, long *result)
{
    int rc;
    
    while(!server->exit_status.should_exit) {
        if(0 != cinvoke_server_handle_request(server)) {
            rc = -1;
            goto Exit;
        }
    }
    
    /* Don't let our exit flag propagate to outer loops */
    server->exit_status.should_exit = CINVOKE_FALSE;
    
    (*result) = server->exit_status.result;
    rc = 0;

Exit:
    return rc;
}

int cinvoke_server_loop(cinvoke_server_t *server, long *result,
                        const void *loop_data, unsigned long loop_data_size)
{
    int rc;
    frame_type_t frame_type = FRAME_TYPE_ENTERING_LOOP;

    start_using_out_pipe(server);
    
    assert_call(cinvoke_writen(server->out_fd, PS(frame_type)));
    assert_call(write_buffer(server->out_fd, loop_data, loop_data_size));

    stop_using_out_pipe(server);

    server->nest_level += 1;
    rc = server_loop(server, result);
    server->nest_level -= 1;
    
    return rc;
}

int cinvoke_server_main_loop(cinvoke_server_t *server, long *result)
{
    return server_loop(server, result);
}

int cinvoke_server_request_handler__EXIT(cinvoke_server_t *server, request_type_t request_type,
                                         cinvoke_bool_t accepted)
{
    long result;
    assert_call(cinvoke_readn(server->in_fd, PS(result)));
    if(!accepted) return 0;
    
    server->exit_status.result = result;
    server->exit_status.should_exit = CINVOKE_TRUE;
    return 0;
}

typedef struct {
    generic_func_ptr_t function_ptr;
    unsigned long extra_size;
    unsigned long stack_size;
    cinvoke_bool_t with_response;
} call_header_t;

static char *allocate_call_buffer(cinvoke_server_t *server, unsigned long buffer_size)
{
    char *allocated_buf;
    assert(server->call_buffer_offset + buffer_size <= BUFFER_SIZE);
    allocated_buf = server->call_buffer + server->call_buffer_offset;
    server->call_buffer_offset += buffer_size;
    return allocated_buf;
}

static void deallocate_call_buffer(cinvoke_server_t *server, unsigned long buffer_size)
{
    server->call_buffer_offset -= buffer_size;
}

int cinvoke_server_request_handler__CALL(cinvoke_server_t *server, request_type_t request_type,
                                         cinvoke_bool_t accepted)
{
    int rc = -1;
    char *buffer_dest;
    char *stack_data;
    call_header_t call_header;
    long result;

    unsigned long total_size;

    /* WARNING: Pay attention to signal reentrancy issues here. */

    assert_call(cinvoke_readn(server->in_fd, PS(call_header)));

    total_size = call_header.extra_size + call_header.stack_size;

    if(!accepted) {
        assert_call(cinvoke_wasten(server->in_fd, total_size));
        rc = 0;
        goto Exit;
    }
    
    buffer_dest = allocate_call_buffer(server, total_size);
    assert(NULL != buffer_dest);

    assert_call(cinvoke_readn(server->in_fd, buffer_dest, total_size));
    
    assert(ALIGNED4(call_header.stack_size));

    stack_data = buffer_dest + call_header.extra_size;

    result = cinvoke_call_func(call_header.function_ptr, stack_data, call_header.stack_size);

    if(call_header.with_response) {
        frame_type_t frame_type = FRAME_TYPE_CALL_RESULT;
        start_using_out_pipe(server);
        assert_call(cinvoke_writen(server->out_fd, PS(frame_type)));
        assert_call(cinvoke_writen(server->out_fd, PS(result)));
        assert_call(cinvoke_writen(server->out_fd, buffer_dest, call_header.extra_size));
        stop_using_out_pipe(server);
    }

    rc = 0;
    
    /* Signal reentrancy: We must deallocate the buffer after we
     * finish using it */
    deallocate_call_buffer(server, total_size);

Exit:
    return rc;
}
