# TODO:
# * debug info of functions for ptr lists

from AddressSpace import AddressSpace
from FunctionPointer import func_ptr_type
from StackArea import StackArea

from DataTypes.Space import attrspace_property
from DataTypes.SubSequence import SubSequence
from DataTypes.MStr import mstr
from DataTypes.BoundFunc import BoundFunc
from DataTypes import CDataTypes

from OSUtils.Executables import exec_info

from MBufProxy import MBufProxy

import struct
import select
import traceback
import weakref
import os
import errno

(FRAME_TYPE_ACCEPT,
 FRAME_TYPE_REJECT,
 FRAME_TYPE_CALL_RESULT,
 FRAME_TYPE_ENTERING_LOOP,
 FRAME_TYPE_EVENT) = xrange(5)

class Error(Exception): pass
class UnknownLoopError(Error): pass
class UnexpectedEvent(Error): pass
class StackError(Error): pass

class LibcErrorResult(Error):
    def __init__(self, str, errno_value):
        self.errno = errno_value
        self.errno_str = errno.errorcode.get(self.errno, "*Unknown*")
        Error.__init__(self, str, self.errno_str)

def unknown_loop_entered(loop_data):
    raise UnknownLoopError("Unknown loop entered", loop_data)

STDIN_FD, STDOUT_FD = xrange(2)

class Client(object):
    EXIT_COMMAND = 0
    CALL_COMMAND = 1
    def __init__(self, exec_info, pipe,
                 event_handler=None,
                 loop_handler=unknown_loop_entered,
                 loop_exception_result=-1,
                 exception_handler=None,
                 loop_exception_handler=None):
        self.exec_info = exec_info
        
        self._stdin, self._stdout = pipe
        self.address_space = AddressSpace(self)
        self.event_handler = event_handler
        
        self.loop_handler = loop_handler
        self.loop_exception_result = loop_exception_result
        self.exception_handler = exception_handler
        self.loop_exception_handler = loop_exception_handler
        
        self._read_buffer_offset()
        
        self._server_stack_mirror = mstr('')
        self._server_stack_mirror_ptr = 0
        self._server_stack_areas = []
        self._server_stack_area_refs = weakref.WeakValueDictionary()
        
        self._heap_buffers = {}

        self.send_queue = []
        self.retransmits = {}
        self.nest_level = 0

    def __del__(self):
        self.close()

    def close(self):
        self._stdout.close()
        self._stdin.close()

    def _send_command(self, command):
        command_data = struct.pack('L', self.nest_level) + command
        self.send_queue.append((self.nest_level, command))
        self._write(command_data)

    def _read_buffer_offset(self):
        self._buffer_offset, = self._read_fmt('L')

    def _write(self, buffer):
        self._stdin.write(buffer)

    def _read(self, size):
        if size == 0:
            return ''
        data = self._stdout.read(size)
        if len(data) == 0:
            raise EOFError()
        assert len(data) == size, len(data)
        return data

    def _read_fmt(self, fmt):
        data = self._read(struct.calcsize(fmt))
        return struct.unpack(fmt, data)

    def _read_buffer(self):
        buffer_length, = self._read_fmt('L')
        data = self._read(buffer_length)
        return data

    def exit(self, result):
        exit_cmd = struct.pack('LL', self.EXIT_COMMAND, result)
        self._send_command(exit_cmd)

    def _handle_event(self):
        event = self._read_buffer()
        if self.event_handler is None:
            raise UnexpectedEvent(repr(event))
        self.event_handler(event)

    def _handle_loop(self, loop_data):
        try:
            result = self.loop_handler(loop_data)
        except:
            self._handle_exception_no_raise(self.loop_exception_handler or self.exception_handler)
            result = self.loop_exception_result
        self.exit(result)

    def _handle_entering_loop(self):
        loop_data = self._read_buffer()
        self.nest_level += 1
        try:
            self._handle_loop(loop_data)
        finally:
            self.nest_level -= 1
        for command in self.retransmits.pop(self.nest_level, []):
            self._send_command(command)

    def _handle_call_acceptance(self, accepted):
        nest_level, command = self.send_queue.pop(0)
        if not accepted:
            assert nest_level <= self.nest_level, nest_level
            if nest_level == self.nest_level:
                self._send_command(command)
            else:
                self.retransmits.setdefault(nest_level, []).append(command)

    _frame_type_handlers = {
        FRAME_TYPE_EVENT : _handle_event,
        FRAME_TYPE_ENTERING_LOOP : _handle_entering_loop,
        FRAME_TYPE_ACCEPT : BoundFunc(_handle_call_acceptance, accepted=True),
        FRAME_TYPE_REJECT : BoundFunc(_handle_call_acceptance, accepted=False),
    }

    def _handle_exception_no_raise(self, exception_handler):
        try:
            if exception_handler is not None:
                exception_handler()
                return
        except:
            pass
        # We want to get here and print the original exception if there is no
        # handler and print the handler's exception if there is a handler and
        # it had an exception.
        
        try:
            traceback.print_exc()
        except:
            # If we can't write exceptions to stderr, we must not raise this,
            # or the protocol gets out of sync!
            pass

    def _read_frame(self):
        frame_type, = self._read_fmt('L')
        if frame_type in self._frame_type_handlers:
            handler = self._frame_type_handlers[frame_type]
            try:
                handler(self)
            except:
                self._handle_exception_no_raise(self.exception_handler)
        else:
            return frame_type

    def _read_next_normal_frame(self):
        while True:
            frame_type = self._read_frame()
            if frame_type is not None:
                return frame_type

    def _read_call_buffer_ptr(self):
        frame_type = self._read_next_normal_frame()
        assert frame_type == FRAME_TYPE_CALL_BUFFER_PTR, frame_type
        buf_offset, = self._read_fmt('L')
        return buf_offset

    def _read_result(self, length):
        frame_type = self._read_next_normal_frame()
        assert frame_type == FRAME_TYPE_CALL_RESULT, frame_type
        result_num, = self._read_fmt('l')
        extra_data = self._read(length)
        return result_num, extra_data
    
    def _do_call(self, func_ptr, extra_data, stack, with_response):
        call_data = struct.pack('LLLLL', self.CALL_COMMAND, func_ptr, len(extra_data), len(stack), with_response)

        call_request = ''.join([call_data, extra_data, stack])
        self._send_command(call_request)
        if with_response:
            result_num, extra_data = self._read_result(len(extra_data))
            return result_num, extra_data

    def _convert_arg(self, node, stack_allocations):
        if isinstance(node, CDataTypes.Node):
            if node.size() < 4:
                raise StackError("C does not support passing arguments under the size of a long on the stack. Use its integer value instead.")
            return node
        elif isinstance(node, (int, long)):
            try:
                return CDataTypes.SignedLong.from_value(mstr, node)
            except OverflowError:
                return CDataTypes.UnsignedLong.from_value(mstr, node)
        elif isinstance(node, str):
            # Allocate the string on the stack, keep it there via
            # the stack_allocations list
            string_allocation = CDataTypes.string_value(self.stack, node)
            stack_allocations.append(string_allocation)
            return CDataTypes.ptr_to(mstr, string_allocation)
        elif isinstance(node, float):
            return CDataTypes.Double.from_value(mstr, node)
        else:
            return node

    def _pack_args(self, args, stack_allocations):
        return ''.join(self._convert_arg(arg, stack_allocations).mbuf[:]
                       for arg in args)

    def _call(self, func_ptr, args, with_response):
        self._free_unused_stack()

        # Keep a list of stack allocations so they stay alive as long
        # until the call completes
        stack_allocations = []
        stack = self._pack_args(args, stack_allocations)
        extra_data = self._server_stack_mirror[self._server_stack_mirror_ptr:]
        total_size = len(extra_data) + len(stack)

        stack_space = self.allocate_stack_space(stack)
        self._server_stack_mirror_ptr += total_size

        try:
            result = self._do_call(func_ptr, extra_data, stack, with_response)
        finally:
            self._server_stack_mirror_ptr -= total_size
        # Once no references to stack_space are left (we hold the only one)
        # it is marked for deallocation, asap
        del stack_space
        del stack_allocations

        result_num = None
        if result is not None:
            result_num, extra_data = result
            self._server_stack_mirror[self._server_stack_mirror_ptr:] = extra_data
        
        # Retry to free stack space, because our call-stack/stack_allocations
        # have probably become freeable
        self._free_unused_stack()
        return result_num


    def _deallocate_stack(self, length):
        self._server_stack_mirror = self._server_stack_mirror[:-len(stack)]

    def allocate_stack_space(self, data):
        self._free_unused_stack()
        
        length = len(data)
        old_len = len(self._server_stack_mirror)
        self._server_stack_mirror += data

        base_stack_offset = self._buffer_offset + old_len
        server_stack_slice = slice(base_stack_offset,
                                   base_stack_offset + len(data))
        server_stack = SubSequence(self.address_space, server_stack_slice)

        client_stack_slice = slice(old_len, old_len + len(data))
        client_stack = SubSequence(self._server_stack_mirror, client_stack_slice)
        def get_current_server_client_barrier():
            # Return a relative offset in the subsequence that is the current
            # barrier between the server and client ownership
            return min(max(0, self._server_stack_mirror_ptr - old_len), len(data))
        # The address space here is the Server address_space, which does not
        # really "contain" this stack area yet. Fortunatly, every read/write to
        # the address space or _any_ other access to the server is done via a
        # "call" and thus this stack area is as good as already there, and
        # addressable via the normal client addressable.
        stack_area = StackArea(base_stack_offset, self.address_space,
                               server_stack, client_stack,
                               get_current_server_client_barrier)
        self._server_stack_area_refs[old_len] = stack_area
        self._server_stack_areas.append(old_len)
        return stack_area

    # The following is useful as an mbuf type
    stack = allocate_stack_space

    def allocate_heap_space(self, data):
        offset = self.functions.malloc(len(data)).value
        subsequence = self.address_range((offset, len(data)))
        subsequence[:] = data
        def free_allocation(x):
            # TODO: Do we need to queue this so it doesn't interfere in the middle of a request?
            self.functions_no_response.free(offset)
            del self._heap_buffers[id(wref)]
        wref = weakref.ref(subsequence, free_allocation)
        self._heap_buffers[id(wref)] = wref
        # Return a MBufProxy because subsequences can be "thrown away" for
        # alternate subsequences at will. Here the actual reference to the
        # subsequence is meaningful. MBufProxy cannot be thrown away
        return MBufProxy(subsequence)

    # The following is useful as an mbuf type
    heap = allocate_heap_space

    def _free_unused_stack(self):
        while True:
            if not self._server_stack_areas:
                # No stack allocated left
                break
            if self._server_stack_areas[-1] in self._server_stack_area_refs:
                # Stack is held alive by a strong reference
                break
            last_stack_offset = self._server_stack_areas.pop()
            del self._server_stack_mirror[last_stack_offset:]
    
    @classmethod
    def from_local_executable(cls, exec_path, **kw):
        info = exec_info(exec_path)
        if not os.path.isabs(exec_path):
            exec_path = './' + exec_path
        pipe = os.popen2(exec_path, bufsize=0)
        
        client = cls(info, pipe, **kw)

        # Its not healthy to stay on stdin/stdout. Some accidentally write on
        # us.
        client.migrate_fds()

        return client

    def environment(self):
        """Return an environment dict suitable as an argument to the
        instantiation of new node types"""
        return dict(address_space = self.address_space)

    def migrate_fds(self, new_input_fd=None, new_output_fd=None):
        """Migrate the fds of the server from input_fd, output_fd to new fds
        
        Assumes that the server is actually operating with INPUT_FD and OUTPUT_FD
        (the cinvoke std main server, and that migrate was not called
        before)"""

        stdmain_server = self.variables.cinvoke_stdmain_server
        input_fd, output_fd = stdmain_server.value.in_fd, stdmain_server.value.out_fd
        if new_input_fd is None:
            new_input_fd = self.libc_success_verifier.dup(input_fd)
        if new_output_fd is None:
            new_output_fd = self.libc_success_verifier.dup(output_fd)
        self.functions.cinvoke_server_migrate_fds(stdmain_server.offset, new_input_fd, new_output_fd)
        self.functions.close(input_fd)
        self.functions.close(output_fd)
    
    def string_of_cstring(self, char_ptr_value, **kw):
        length = self.functions.strlen(char_ptr_value).value
        string_buf_type = CDataTypes.buffer_size(length+1)
        return self.node_at(string_buf_type, char_ptr_value, **kw)
    
    def string_array(self, strings):
        from warnings import warn
        warn(DeprecationWarning("The Client.string_array method is obselete. Use the CDataTypes.string_array function instead."))
        return CDataTypes.string_array(self.stack, strings)

    def address_range(self, (offset, size)):
        return SubSequence(self.address_space,
                           slice(offset, offset+size))

    def get_symbol_range(self, symbol_name):
        symbol_area = self.exec_info.symbol_area(symbol_name)
        return self.address_range(symbol_area)
    symbol_ranges = attrspace_property(get_symbol_range)

    def node_at(self, node_type, offset, environment=None, **kw):
        if environment is None:
            environment = self.environment()
        subsequence = self.address_range((offset, node_type.size()))
        return node_type(subsequence, environment, **kw)

    def node_symbol(self, node_type, name, **kw):
        offset, size = self.exec_info.symbol_area(name)
        if size != node_type.size():
            raise ValueError("Symbol %r is of size %d, not %d" % (name, size, node_type.size()))
        return self.node_at(node_type, offset, **kw)
    
    def node_type(self, node_type):
        class ClientNodeType(node_type):
            def __init__(cp_self, *args, **kw):
                environment = kw.get('environment')
                if environment is None:
                    kw['environment'] = self.environment()
                super(ClientNodeType, cp_self).__init__(*args, **kw)
        ClientNodeType.__name__ = 'Client' + node_type.__name__
        return ClientNodeType

    def __get_libc_verify_function(self, func_name):
        func = getattr(self.functions, func_name)
        def verify_call(*args, **kw):
            result = func(*args, **kw)
            if -1 == result.value:
                errno_value = self.variables.errno.value
                raise LibcErrorResult("Call to %r failed" % (func_name,), errno_value)
            return result
        return verify_call
    libc_success_verifier = attrspace_property(__get_libc_verify_function)

    def _caller_by_name(self, mbuf_type, name, with_response, **kw):
        fptr_type = func_ptr_type(**kw)
        fptr = fptr_type.from_zero(mbuf_type, client=self, with_response=with_response)
        fptr.value_name = name
        return fptr

    def __caller_by_name_with_response(self, name):
        return self._caller_by_name(mstr, name, True)

    def __caller_by_name_without_response(self, name):
        return self._caller_by_name(mstr, name, False)
    
    functions = attrspace_property(__caller_by_name_with_response, doc="Callable functions namespace")
    functions_no_response = attrspace_property(__caller_by_name_without_response, doc="Callable functions namespace (no response - use if the function has no output parameters or return value)")

    def __get_var_by_name(self, name):
        node_type = getattr(self.exec_info.variables, name).type
        return self.node_symbol(node_type, name)
    variables = attrspace_property(__get_var_by_name)

    def handle_data(self):
        """Handle incoming events, if any are available"""
        while True:
            r, w, x = select.select([self._stdout], [], [], 0)
            if not r:
                # No data available
                break
            self._handle_next_frame()

    def _handle_next_frame(self):
        assert self._read_frame() is None, "Unexpected synchronous frame arrived"
