"""A list-like or tuple-like object, in which the fields are also accessible by
name.

Examples:
>>> from DataTypes.AttrSeq import AttrTuple, AttrList

>>> b=AttrTuple([('a', 1), ('b', 2)])
>>> b
(a = 1, b = 2)
>>> b.a
1
>>> b.b=2
Traceback (most recent call last):
  File "...", line ..., in ...
TypeError: object does not support item assignment
>>> b[0]
1

>>> b=AttrList([('a', 1), ('b', 2)])
>>> b
[a = 1, b = 2]
>>> b.b=3
>>> b
[a = 1, b = 3]
"""

private_prefix = '_AttrSeq'
class AttrSeq(object):
    seq_type = None
    parens = None
    def __init__(self, fields):
        if isinstance(fields, AttrSeq):
            fields = zip(fields.__names, fields.__values)
        # We iterate it twice, so make sure we get the same thing...
        fields = list(fields)
        
        self.__names = [name for name, value in fields]
        if len(set(self.__names)) != len(self.__names):
            raise ValueError("Names not unique", fields)
        self.__values = self.seq_type(value for name, value in fields)

    def __repr__(self):
        start_paren, end_paren = self.parens
        args_str = ', '.join('%s = %r' % (name, value) for name, value in zip(self.__names, self.__values))
        return '%s%s%s' % (start_paren, args_str, end_paren)

    def __hash__(self):
        return hash(self.__values)

    def __len__(self):
        assert len(self.__names) == len(self.__values)
        return len(self.__names)

    def __cmp__(self, other):
        return cmp(self.__values, other)

    def __index_of_name(self, name):
        if name not in self.__names:
            raise AttributeError("No such attribute", name)
        return self.__names.index(name)
    
    def __getattr__(self, name):
        if name.startswith(private_prefix):
            return super(AttrSeq, self).__getattr__(name)
        else:
            index = self.__index_of_name(name)
            return self.__values[index]
    def __setattr__(self, name, value):
        if name.startswith(private_prefix):
            super(AttrSeq, self).__setattr__(name, value)
        else:
            index = self.__index_of_name(name)
            self.__values[index] = value
    def __delattr__(self, name):
        index = self.__index_of_name(name)
        del self.__values[index]
        del self.__names[index]
    
    def __getitem__(self, index):
        return self.__values[index]
    def __setitem__(self, index, value):
        cur_values = self.__values[index]
        try:
            value_len = len(value)
            values_len = len(cur_values)
        except TypeError:
            pass
        else:
            if value_len != values_len:
                raise ValueError("Cannot resize via item assignment")
        self.__values[index] = value
    def __delitem__(self, index):
        del self.__values[index]

class AttrTuple(AttrSeq):
    seq_type = tuple
    parens = '()'

class AttrList(AttrSeq):
    seq_type = list
    parens = '[]'
