from DictUtils import sum_dicts

class MultiDirBoundFunc(object):
    """Create a function with partially bound arguments, that can be placed both before and after the arguments of the call.
    Examples:
    >>> from DataTypes.BoundFunc import MultiDirBoundFunc
    >>> def joe(a,b): print a,b
    ... 
    >>> MultiDirBoundFunc(joe, [], [2])(1)
    1 2
    >>> MultiDirBoundFunc(joe, [2], [])(1)
    2 1
    """
    def __init__(self, func, pre_args=(), post_args=(), kw=None):
        if not callable(func):
            raise TypeError('Must use a callable with MultiDirBoundFunc')
        self.func = func
        self.pre_args = pre_args
        self.post_args = post_args
        if kw is None:
            kw = {}
        self.kw = kw
    def __repr__(self):
        def reprs(items):
            return [repr(item) for item in items]
        def kw_reprs(kw_dict):
            return ['%s=%r' % item
                    for item in kw_dict.iteritems()]
        args_str = ', '.join(reprs(self.pre_args) + ['...'] + reprs(self.post_args) + list(kw_reprs(self.kw)))
        return '<%r bound to (%s)>' % (self.func, args_str)
    def __call__(self, *more_args, **more_kw):
        args = tuple(self.pre_args) + more_args + tuple(self.post_args)
        return self.func(*args, **sum_dicts(self.kw, more_kw,
                                            allow_override=False))

class BoundFunc(MultiDirBoundFunc):
    """Blackwards-Compatibility Layer for MultiDirBoundFunc
    Create a function with partially bound arguments.

    Examples:
    >>> def f(a, b, c, d=0):
    ...   print a, b, c, d
    ... 
    >>> from DataTypes.BoundFunc import BoundFunc
    >>> b=BoundFunc(f, 1, d=2)
    >>> b(3,4)
    1 3 4 2
    >>> b(3,4,d=5)
    Traceback (most recent call last):
      File "...", line ..., in ...
    ValueError: Overriding dict values disallowed
    >>> b=BoundFunc(f, 1, 2)
    >>> b(3)
    1 2 3 0
    >>> b(3, 4)
    1 2 3 4
    """
    def __init__(self, func, *args, **kw):
        super(BoundFunc, self).__init__(func, args, (), kw)
