__all__ = ["Bunch", "attributes_of"]

def repr_lock_wrapper(func):
    def wrapper(self, *args, **kw):
        # TODO: store in "thread local storage"
        if getattr(self, '_Bunch__in_repr', False):
            return '%s(...)' % (self.__class__.__name__,)
        try:
            setattr(self, '_Bunch__in_repr', True)
            ret = func(self, *args, **kw)
        finally:
            setattr(self, '_Bunch__in_repr', False)
        return ret
    return wrapper

def attributes_of(bunch):
    for name in bunch.field_order:
        yield name, getattr(bunch, name)
    for name, value in sorted(bunch.__dict__.iteritems()):
        if name in bunch.field_order:
            continue
        if name.startswith('_Bunch__'):
            continue
        yield name, value


class Bunch(object):
    """Bunch up a bunch of attributes_of.

    Examples:
    >>> from DataTypes.Bunch import Bunch
    >>> car = Bunch(speed = 1, fuel_tank_size = 20)
    >>> car
    Bunch(fuel_tank_size = 20, speed = 1)
    >>> car.fuel_tank_size
    20
    >>> car.speed*=5
    >>> car
    Bunch(fuel_tank_size = 20, speed = 5)
    >>> class Car(Bunch):
    ...   # Calling the super-constructor is optional
    ...   def __init__(self):
    ...     self.speed = 10
    ...     self.fuel_tank_size = 20
    ... 
    >>> c=Car()
    >>> c
    Car(fuel_tank_size = 20, speed = 10)
    >>> c.speed
    10
    >>> recursive = Bunch()
    >>> recursive.recursive = recursive
    >>> recursive
    Bunch(recursive = Bunch(...))
    """
    field_order = []
    
    def __init__(self, **kw):
        self.__dict__.update(kw)

    @repr_lock_wrapper
    def __repr__(self):
        attrs = ('%s = %r' % item for item in attributes_of(self))
        return '%s(%s)' % (self.__class__.__name__, ', '.join(attrs))

    def __cmp__(self, other):
        return (cmp(type(self), type(other)) or 
                cmp(list(attributes_of(self)),
                    list(attributes_of(other))))

    # Pretty printing stuff:

    INDENT_WITH = "    "
    
    # TODO: This should be in a pretty-print (a.k.a pp) module
    def __format_seq(self, seq, nesting):
        indentation = self.INDENT_WITH * nesting

        # is there any Bunch contained in the sequence?
        for elem in seq:
            if isinstance(elem, Bunch):
                break
        else:
            # no - just repr it
            return repr(seq)
        
        # yes - make it indented
        text = ['[']
        for elem in seq:
            text.append('\n')
            text.append(indentation)
            if isinstance(elem, Bunch):
                text.append(elem.pretty_str(nesting + 1))
            else:
                text.append(repr(elem))
            text.append(',')
        text.append('\n')
        text.append(self.INDENT_WITH * (nesting - 1))
        text.append(']')
        return ''.join(text)
    
    @repr_lock_wrapper
    def pretty_str(self, nesting = 1):
        indentation = self.INDENT_WITH * nesting
        attrs = []
        for name, value in attributes_of(self):
            if isinstance(value, Bunch):
                attrs.append('%s = %s' % (name, value.pretty_str(nesting + 1)))
            elif isinstance(value, list) or isinstance(value, tuple):
                attrs.append('%s = %s' % (name, self.__format_seq(value, nesting + 1)))
            else:
                attrs.append('%s = %r' % (name, value))
        if not attrs:
            attrs = ["<empty>"]
        attrs = ['%s%s' % (indentation, a) for a in attrs]
        return '%s:\n%s' % (self.__class__.__name__, '\n'.join(attrs).rstrip())
    
