from DataTypes.FuncCache import func_cache
from DataTypes.SequenceUtils import chunks
from DataTypes.Space import itemspace_property
from DataTypes.SubSequence import SubSequence
from DataTypes.SliceUtils import slice_indices
from DataTypes.CDataTypes.FormatNodes import string_value, Char
from DataTypes.CDataTypes.Pointers import ptr_type
from Node import Node

class Array(Node):
    node_type = None
    count = None

    @classmethod
    def is_abstract(cls):
        return cls.count is None
        
    def __len__(cls):
        return cls.count

    def _mbuf_at(self, index):
        size = self.node_type.size()
        return SubSequence(self.mbuf, slice(index*size, (1+index)*size))
    
    def _get_value(self, index):
        return self.node_type.unpack(self._mbuf_at(index)[:])

    def _get_node(self, index):
        return self.node_type(self._mbuf_at(index), environment=self.environment)

    def _set_item(self, index, value):
        size = self.node_type.size()
        self.mbuf[index*size:(1+index)*size] = self.node_type.pack(value)

    def _convert_index(self, index):
        if index < 0:
            index += len(self)
        if index < 0 or (self.count is not None and index >= self.count):
            raise IndexError("Array index out of bounds")
        return index

    def _xrange(self, slice):
        start, stop, step = slice.indices(len(self))
        return xrange(start, stop, step)

    def __get_values(self, index):
        if isinstance(index, slice):
            return tuple(self._get_value(i) for i in self._xrange(index))
        return self._get_value(self._convert_index(index))
    
    def __set_values(self, index, value):
        if isinstance(index, slice):
            xr = self._xrange(index)
            if len(value) != len(xr):
                raise ValueError("Cannot resize Arrays via slice assignment")
            for i, item in zip(xr, value):
                self._set_item(i, item)
        else:
            self._set_item(self._convert_index(index), value)

    def __get_nodes(self, index):
        if isinstance(index, slice):
            return tuple(self._get_node(i) for i in self._xrange(index))
        return self._get_node(self._convert_index(index))

    def __get_subarray(self, index):
        if isinstance(index, slice):
            index = slice_indices(index, len(self))
            assert index.step in [None, 1], "step must be 1"
            new_array_type = array_of(self.node_type, (index.stop-index.start))
            size = self.node_type.size()
            mbuf_index = slice(index.start * size,
                               index.stop * size)
            return new_array_type(SubSequence(self.mbuf, mbuf_index),
                                  environment=self.environment)
        return self._get_node(self._convert_index(index))
    
    # If you are not accessing the whole array at a time, then this is better
    values = itemspace_property(__get_values, __set_values,
                                doc="Use this attribute to read/write some of the values")

    nodes = itemspace_property(__get_nodes,
                               doc="Use this attribute to read some of the nodes")

    subarray = itemspace_property(__get_subarray,
                                  doc="Use this attribute to get a subarray node")

    @classmethod
    def size(cls):
        return cls.node_type.size() * cls.count

    @classmethod
    def _pack(cls, value):
        if len(value) > cls.count:
            raise PackError("Array requires at most %d elements, not %d" % (cls.count, len(value)))
        packed_values = ''.join(cls.node_type._pack(item) for item in value)
        return packed_values + '\x00' * cls.node_type.size() * (cls.count - len(value))

    @classmethod
    def _unpack(cls, buffer):
        return tuple(cls.node_type._unpack(chunk)
                     for chunk in chunks(buffer, cls.node_type.size()))

    @classmethod
    def align_size(cls):
        return cls.node_type.align_size()

    def __str__(self):
        return ''.join(map(chr, self.value))

    def __unicode__(self):
        return ''.join(map(unichr, self.value))

@func_cache
def abstract_array(element_type):
    class TempArray(Array):
        node_type = element_type
    TempArray.__name__ = 'AbstractArray_%s' % (element_type.__name__,)
    return TempArray

def array_of(element_type, element_count):
    if not isinstance(element_type, type) or not issubclass(element_type, Node):
        raise TypeError("Node Type required but %r given" % (element_type,))
    class TempArray(abstract_array(element_type)):
        node_type = element_type
        count = element_count
    TempArray.__name__ = 'Array_%d_%s' % (element_count, element_type.__name__)
    return TempArray

# Python string -> C string array..
# Ugly.. any way to generalize on this concept?
def string_array(mbuf_type, strings):
    allocated_strings = [string_value(mbuf_type, string)
                         for string in strings]
    ptrs = [allocated_string.offset
            for allocated_string in allocated_strings]

    ptr_array_type = array_of(ptr_type(Char), len(strings))
    ptr_array = ptr_array_type.from_value(mbuf_type, ptrs)
    ptr_array.refs_to = allocated_strings
    return ptr_array
