from DataTypes.AttrSeq import AttrTuple
from DataTypes.Space import attrspace_property
from DataTypes.SubSequence import SubSequence
from Errors import PackError
from Node import Node

def align_offset(offset, size):
    return ((size - offset) % size) + offset

class _FieldContainer(Node):
    field_types = None
    
    @classmethod
    def _index_of_name(cls, name, error):
        for index, (field_name, field_type) in enumerate(cls.field_types):
            if name == field_name:
                return index
        raise error

    def __get_node_by_name(self, name):
        index = self._index_of_name(name, AttributeError("No such node", name))
        name, (start, stop), field_type = list(self.field_positions())[index]
        subsequence = SubSequence(self.mbuf, slice(start, stop))
        return field_type(subsequence, self.environment)

    nodes = attrspace_property(__get_node_by_name)

    @classmethod
    def align_size(cls):
        size = 1
        for name, field_type in cls.field_types:
            size = max(size, field_type.align_size())
        return size

class Union(_FieldContainer):
    @classmethod
    def field_positions(cls):
        for name, field_type in cls.field_types:
            yield name, (0, field_type.size()), field_type

    @classmethod
    def _pack(cls, value):
        return value
    
    @classmethod
    def _unpack(cls, buffer):
        return buffer

    @classmethod
    def size(cls):
        if not cls.field_types:
            return 0
        return max(max(field_type.size() for name, field_type in cls.field_types), 0)

class Struct(_FieldContainer):
    packing_level = 1
    
    @classmethod
    def field_positions(cls):
        start = 0
        for name, field_type in cls.field_types:
            align_size = field_type.align_size()
            size = field_type.size()
            start = align_offset(start, min(cls.packing_level, align_size))
            yield name, (start, start+size), field_type
            start += size

    @classmethod
    def field_postpadding(cls):
        if not cls.field_types:
            return
        field_iter = iter(cls.field_positions())
        name, (start, stop), field_type = field_iter.next()
        prev_stop = stop
        for name, (start, stop), field_type in field_iter:
            yield start-prev_stop
            prev_stop = stop
        yield cls.size() - prev_stop

    @classmethod
    def size(cls):
        positions = list(cls.field_positions())
        if not positions:
            return 0
        name, (start, stop), field_type = positions[-1]
        return align_offset(stop, cls.packing_level)
    
    @classmethod
    def _pack_field_value(cls, field_type, postpad, value):
        if isinstance(value, Node):
            raise PackError("Must use values, not nodes")
        return field_type.pack(value) + '\x00'*postpad

    @classmethod
    def _pack_dict(cls, values):
        for postpad, (name, field_type) in zip(cls.field_postpadding(), cls.field_types):
            value = values[name]
            yield cls._pack_field_value(field_type, postpad, value)

    @classmethod
    def _pack_sequence(cls, values):
        for postpad, (name, field_type), value in zip(cls.field_postpadding(), cls.field_types, values):
            yield cls._pack_field_value(field_type, postpad, value)
    
    @classmethod
    def _pack(cls, value):
        if len(value) != len(cls.field_types):
            raise PackError('Attempt to pack %d fields, %d fields required' %
                            (len(value), len(cls.field_types)))
        if isinstance(value, dict):
            packed_fields = cls._pack_dict(value)
        else:
            packed_fields = cls._pack_sequence(value)
        return ''.join(packed_fields)
    
    @classmethod
    def _unpack(cls, buffer):
        return AttrTuple(
            (name, field_type.unpack(buffer[start:stop]))
            for name, (start, stop), field_type in cls.field_positions()
        )
