from Node import Node
from DataTypes.FuncCache import func_cache
import struct

__all__ = [
    # The basic type combinations (signed/unsigned littleendian/bigendian
    # short/int/long/etc are auto-generated below)
    "Char",
    "SignedChar",
    "UnsignedChar",
    "Float",
    "Double",
    "buffer_size",
    "buffer_value",
    "string_value",
]

class _FmtNode(Node):
    format = None
    
    @classmethod
    def size(cls):
        return struct.calcsize(cls.format)
    
    @classmethod
    def _pack(cls, value):
        if cls.format.isupper() and value < 0:
            raise ValueError('Attempt to pack negative value %r in %r, only positive values allowed' %
                             (value, cls))
        return struct.pack(cls.format, value)
    
    @classmethod
    def _unpack(cls, buffer):
        value, = struct.unpack(cls.format, buffer)
        return value

    @classmethod
    def align_size(cls):
        return cls.size()

def define_basic_types():
    for type_string, type_fmt in (('',         'i'), # default is int
                                  ('Int',      'i'),
                                  ('Long',     'l'),
                                  ('LongLong', 'q'),
                                  ('Short',    'h')):
        for signed_string, signed in (('', True), # default is signed
                                      ('Signed',   True),
                                      ('Unsigned', False)):
            for endian_string, endianity in (('' , ''), # default is platform default
                                             ('BigEndian', '>'),
                                             ('LittleEndian', '<')):

                partial_name = ''.join((signed_string, type_string))
                if not partial_name:
                    continue

                name = ''.join((endian_string, partial_name))

                if signed:
                    fmt = endianity + type_fmt.lower()
                else:
                    fmt = endianity + type_fmt.upper()

                class TempFormatType(_FmtNode):
                    format = fmt

                TempFormatType.__name__ = name
                globals()[name] = TempFormatType
                __all__.append(name)

define_basic_types()

class _BaseChar(_FmtNode):
    @classmethod
    def _pack(cls, value):
        try:
            value = ord(value)
        except TypeError:
            pass
        return super(_BaseChar, cls)._pack(value)

class Char(_BaseChar): format = 'b'
SignedChar = Char
class UnsignedChar(_BaseChar): format = 'B'

class Float(_FmtNode): format = 'f'

class Double(_FmtNode): format = 'd'

class _FmtBuffer(_FmtNode):
    @classmethod
    def align_size(cls):
        return 1

@func_cache
def buffer_size(x):
    class Buffer(_FmtBuffer):
        format = '%ds' % (x,)
    Buffer.__name__ = 'Buffer%dBytes' % (x,)
    return Buffer

def buffer_value(mbuf_type, value, allocation_size=None, **kw):
    if allocation_size is None:
        allocation_size = len(value)
    return buffer_size(allocation_size).from_value(mbuf_type, value, **kw)

def string_value(mbuf_type, value, **kw):
    return buffer_value(mbuf_type, value + '\x00', **kw)
