from DataTypes.SubSequence import SubSequence
from DataTypes.TypeUtils import classname
from Errors import PackError, UnpackError

class Node(object):
    def __init__(self, mbuf, environment=None):
        super(Node, self).__init__()
        if environment is None:
            environment = {}
        if not self.is_abstract():
            if len(mbuf) != self.size():
                raise ValueError("SubSequence of length %r given to %s of size %d" %
                                 (len(mbuf), classname(self), self.size()))
        self._mbuf = mbuf
        self.environment = environment

    # Make mbuf a read-only reference
    def __get_mbuf(self):
        return self._mbuf
    mbuf = property(__get_mbuf)

    def __get_offset(self):
        return self.mbuf.offset
    offset = property(__get_offset)
    
    def __get_address_space(self):
        return self.mbuf.address_space
    address_space = property(__get_address_space)
    
    def copy(self, mbuf_type, environment=None, **kw):
        mbuf = mbuf_type(self.mbuf)
        if environment is None:
            environment = self.environment
        node_type = type(self)
        new_node = node_type(mbuf, environment, **kw)
        return new_node

    def cast(self, node_type):
        return node_type(self.mbuf, self.environment)

    @classmethod
    def is_abstract(cls):
        """Returns whether this node is abstract.

        Abstract nodes are nodes that have no size or a complete
        value.  Some abstract nodes can offer some functionality other
        than a size or value, such as accessing parts of the value.
        """
        return False

    def __repr__(self):
        return '%s(%s)' % (classname(self), ', '.join(self._repr_fields()))

    def _repr_fields(self):
        return ['value=%r' % (self.value,)]

    def __hash__(self):
        raise TypeError("%r objects are mutable and thus unhashable" % (classname(self),))

    def __cmp__(self, other):
        return (cmp(type(self), type(other)) or
                cmp(self.value, other.value))

    @classmethod
    def zerobuf(cls):
        return '\x00' * cls.size()

    @classmethod
    def from_zero(cls, mbuf_type, **kw):
        return cls(mbuf_type(cls.zerobuf()), **kw)
    
    def __get_value(self):
        return self.unpack(self.mbuf[:])
    def __set_value(self, value):
        self.mbuf[:] = self.pack(value)
    value = property(__get_value, __set_value)


    @classmethod
    def pack(cls, value):
        """Pack the given value to a buffer according to the format of this node"""
        result = cls._pack(value)
        if len(result) != cls.size():
            raise PackError("Pack returned %d bytes, not %d bytes" % (len(result), cls.size()))
        return result

    @classmethod
    def unpack(cls, buffer):
        """Unpack the given buffer to a value according to the format of this node"""
        if len(buffer) != cls.size():
            raise UnpackError("Unpack given %d bytes, not %d bytes" % (len(buffer), cls.size()))
        return cls._unpack(buffer)

    @classmethod
    def from_value(cls, mbuf_type, value, **kw):
        s = mbuf_type(cls.pack(value))
        return cls(s, **kw)

    @classmethod
    def size(cls):
        raise NotImplementedError()
    
    @classmethod
    def _pack(cls, value):
        raise NotImplementedError()
    
    @classmethod
    def _unpack(cls, buffer):
        raise NotImplementedError()

    @classmethod
    def align_size(cls):
        """gcc has a horrible packing algorithm which packs fields in struct
        according to their type.  Instead of using their type size directly, it
        uses the type size in a manner which depends on the node type.  For
        example, the size of basic types is used directly, while the size of
        arrays uses the size of each element.
        """
        raise NotImplementedError()

