r"""C datatypes package.

This module "mirrors" the C data type system, and allows the creation of
arrays, structs, unions, basic types (int, long, double, etc), enums, and even
pointers.

It is a generic type library, not to be confused with the CInvoke
client, which utilizes it for its own evil gain.

The Data Types library consists of several key concepts:

The mbuf interface: Basically a mutable string object (read/write via slice
indexing).  mbuf's also optionally include an "offset" and "address_space"
fields, if they can be pointed to.

Node Types: These are all *CLASSES* (not instances!) which are subclasses of
the base Node class. All Node Types represent some C Data Type of a
*FIXED-SIZE* (just like all types in C have fixed sizes).  For example:

>>> from DataTypes import CDataTypes
>>> CDataTypes.UnsignedLong.size()
4
>>> CDataTypes.UnsignedLong.pack(5)
'\x05\x00\x00\x00'
>>> CDataTypes.UnsignedLong.unpack(_)
5L
>>> CDataTypes.BigEndianLongLong.pack(1)
'\x00\x00\x00\x00\x00\x00\x00\x01'

In the above example, the "UnsignedLong" Node Type (which is a CLASS, not an
INSTANCE) is used to convert a VALUE (unpacked representation) to its PACKED
REPRESENTATION as a string.  That's basically what Node Types are good
for. Packing/unpacking values/strings.

Thus:

A Node Type (class): Represents some C data structure.
A value: Represents the unpacked representation of the value (for example, a
         simple Python integer)

Another key concept is the Node instance.  Every above-mentioned Node Type can
be instantiated. The Node instance contains a reference to an mbuf. The mbuf in
the Node instance contains the packed representation of the data, and is
accessible via node.mbuf.  Every Node instance contains a "value" property
which uses the Node Type's unpack method on the mbuf, providing a seamless way
to access the Node's mbuf data in its unpacked representation.

In order to create Node instances, we need to have an mbuf readily
available. Since mbuf is just an interface (a mutable string readable and
writable via slice indexing), we need an actual class that implements that
interface. There are many classes which do this, and we will use the MStr class
which is a simple Mutable String that "lives" in the Python interpreter memory,
as a thin wrapper layer over the builtin array.array type (See the MStr
documentation for more information).

>>> from DataTypes.MStr import mstr
>>> m = mstr('\x05\x00\x00\x00')
>>> x = CDataTypes.UnsignedLong(m)
>>> x.value
5L
>>> x.value = 6
>>> m
mstr('\x06\x00\x00\x00')

The above example creates an instance of UnsignedLong and uses it to manipulate
an mstring via its unpacked value representation.  As you can see, initializing
the above mstr separately from the Node instantiation was not very
convinient. Thus, Node Types (the classes) have alternative methods to
construct instances, which are more convinient: from_zero and from_value.
from_zero and from_value take as their first argument, an mbuf_factory. An
mbuf_factory is just a function that creates the mbuf for the node. For
example, mstr itself is a suitable mbuf factory that creates mstr objects.  You
can use any function that returns an mbuf-compatible object as your
mbuf_factory.  In practice, users of CDataTypes often create mbuf's by using a
factory that returns a SubSequence of some other MStr (see the documentation
for SubSequence).

When we just want to create an instance, and don't really care what value it
contains, we can use from_zero:

>>> from DataTypes.MStr import mstr
>>> x = CDataTypes.UnsignedLongLong.from_zero(mstr)
>>> x
UnsignedLongLong(value=0L)
>>> x.value = 5
>>> x
UnsignedLongLong(value=5L)
>>> from DataTypes.MStr import mstr
>>> x = CDataTypes.UnsignedLongLong.from_zero(mstr)
>>> x
UnsignedLongLong(value=0L)
>>> x.value = 5

As mentioned earlier, we can read the packed representation of the Node by
accessing its mbuf attribute. This is more efficient than "pack"ing the value,
and less silly, because every access to the "value" property actually unpacks
the mbuf. Using node.pack(node.value) actually translates to an unpack and
repack of the node's mbuf.

>>> x.mbuf
mstr('\x05\x00\x00\x00\x00\x00\x00\x00')

To create Struct nodes, you need to subclass CDataTypes.Struct and define the
field_types, and optionally define the packing_level. The packing_level is 1 by
default (if not mentioned), and is the equivalent of gcc's '#pragma pack'
statement.

>>> class MyStruct(CDataTypes.Struct):
...   packing_level = 4 # pragma pack like statement is optional
...   field_types = [
...     ('a', CDataTypes.Short),
...     ('b', CDataTypes.BigEndianLongLong),
...   ]
... 
>>> MyStruct.size()
12

Note how the above struct, composed of a Short (2 bytes), and a
BigEndianLongLong (8 bytes), is 12 bytes long.  This is because of the
packing_level, which causes BigEndianLongLong to be padded by 2 bytes. The
alignment rules of gcc are not very simple, aren't portable and are out of the
scope of this documentation, but are (as far as I know) perfectly immitated by
this library.

Earlier we saw that the value (unpacked representation) of various types
(UnsignedLong, BigEndianLongLong, etc) are Python integers.  In contrast, the
value (unpacked representation) of struct node types can be any sequence of
values (each's type is the one that matches the appropriate contained node
type), OR it can be a dictionary mapping struct field names to the appropriate
value.

A dict value:

>>> MyStruct.pack(dict(a=1,b=2))
'\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x02'

A tuple value:

>>> MyStruct.pack((-1, -1))
'\xff\xff\x00\x00\xff\xff\xff\xff\xff\xff\xff\xff'

When the Struct returns its value, it uses an AttrTuple:

>>> MyStruct.unpack('\xff\xff\x00\x00\xff\xff\xff\xff\xff\xff\xff\xff')
(a = -1, b = -1L)

Defining Unions is almost identical to defining and using structs:

>>> class MyUnion(CDataTypes.Union):
...   field_types = [
...     ('a', CDataTypes.array_of(CDataTypes.Char, 3)),
...     ('b', CDataTypes.UnsignedLong),
...   ]
...

However, the unpacked representation of Unions is exactly the same as their
packed representation (There really is no meaningful way to define an unpacked
value of a union).

>>> MyUnion.pack('1234')
'1234'
>>> MyUnion.unpack('1234')
'1234'

So, if unions cannot unpack strings properly, what ARE they useful for?
Lets see:

>>> mu=MyUnion.from_zero(mstr)
>>> mu
MyUnion(value='\x00\x00\x00\x00')

That wasn't interesting, was it?  That's because a Union is mainly used to
create Node instances that represent the contained fields.  This is done by
using the "nodes" attrspace (See the Space.py documentation) in the Union
instance.  The nodes inside the attrspace return Nodes that map to a
SubSequence of the mbuf of the union, such that they share the space as real C
unions:

>>> mu.nodes.a
Array_3_Char(value=(0, 0, 0))
>>> mu.nodes.b
UnsignedLong(value=0L)
>>> mu.nodes.b.value=0x01020304
>>> mu
MyUnion(value='\x04\x03\x02\x01')
>>> mu.nodes.a
Array_3_Char(value=(4, 3, 2))
>>> mu.nodes.b
UnsignedLong(value=16909060L)

Also note in the above example, that the value (unpacked representation) of an
Array type is a simple tuple. Note how changing nodes.b changed the array.

Enums are created in a similar manner to structs, by subclassing the Enum type,
and defining the value_names attribute, as a sequence of (name, value) tuples:

>>> class Moshe(CDataTypes.Enum):
...   value_names = [
...     ('a', 1),
...     ('b', 2),
...   ]
... 
>>> Moshe.size()
4

The Enum uses the Integer type by default, but unlike C, this is overridable by
specifying the node_type class-attribute in the Enum subclass, to specify the
container Node type of the enum.

The Enum can be used simply:

>>> Moshe.pack(3)
'\x03\x00\x00\x00'
>>> a=Moshe.from_zero(mstr)
>>> a
Moshe(value_name='*Unknown*(0)')

Note how the Enum represents itself via the "value_name" attribute, rather than
the value attribute. Since we haven't defined a name for the 0 value, it is
shown as "*Unknown*(0)"

>>> a.value = 1
>>> a
Moshe(value_name='a')

The "value_name" is not just repr-candy. It can also be used directly to assign
values into the enum:

>>> a.value_name = 'b'
>>> a
Moshe(value_name='b')
>>> a.value
2
"""

# The following names are all imported here for export purposes.
# Feel free to use any of them!
from Errors import PackError, UnpackError, DerefError
from Node import Node
from Typedef import typedef
from FieldContainers import Struct, Union
from NamedValueNode import NamedValueNode
from Pointers import Pointer, ptr_type, ptr_to
from Array import Array, abstract_array, array_of, string_array
from Enum import Enum, enum_on, enum_of
from SymbolPointer import SymbolPointer

# Auto-generated list in FormatNodes makes name listing worse than a horrible *
from FormatNodes import *

def cast(from_type, x, to_type):
    return to_type.unpack(from_type.pack(x))

# TODO:
# * Comparability of Node Types with a metaclass?
# * Split this up to a huge package, instead of a huge module...
# * Access pointers as arrays (Pointer should subclass same class as Array)?
