"""SubSequence: A "living slice" to an object.

A SubSequence is an object which represents a slice area inside another
random-access sequence object.

Reading/writing slices from/to a SubSequence actually reads/writes them from
the original slice.
It is ofcourse possible, to recursively create SubSequences of SubSequences.

Examples:

    >>> a = [1, 2, 3, 4, 5]
    >>> b = SubSequence(a, slice(2, 4))
    >>> b[:]
    [3, 4]
    >>> b[:1]=[5]
    >>> b[1]=7
    >>> a
    [1, 2, 5, 7, 5]
    >>> b[:]
    [5, 7]

"""

from TypeUtils import classname
from SliceUtils import combine_slices, combine_slice_index, slice_len, fixup_slice, normalize_slice

class SubSequence(object):
    def __init__(self, sequence, slice_obj=slice(None,None,None)):
        if isinstance(sequence, SubSequence):
            slice_obj, sequence = (
                combine_slices(sequence._slice_obj, sequence._fixup_slice(slice_obj)),
                sequence.full_sequence
            )

        self.full_sequence = sequence
        self._slice_obj = fixup_slice(slice_obj, self._full_length())
        
    def _full_length(self):
        try:
            return len(self.full_sequence)
        except TypeError:
            return None

    def __get_offset(self):
        return self.full_sequence.offset + self._slice_obj.start
    offset = property(__get_offset)

    def __get_address_space(self):
        return self.full_sequence.address_space
    address_space = property(__get_address_space)

    def __hash__(self):
        raise TypeError("%s objects are unhashable" % (classname(self),))
    def __cmp__(self, other):
        return cmp(self[:], other)
    def __len__(self):
        return slice_len(self._slice_obj)
    def __repr__(self):
        return '<%s object (%s, %r)>' % (classname(self), hex(id(self.full_sequence)), self._slice_obj)
    def _fixup_slice(self, slice_obj):
        return fixup_slice(slice_obj, len(self))
    def __getitem__(self, index):
        if isinstance(index, slice):
            index = combine_slices(self._slice_obj, self._fixup_slice(index))
        else:
            index = combine_slice_index(self._slice_obj, index)
        return self.full_sequence[index]
    def __setitem__(self, index, value):
        if isinstance(index, slice):
            index = combine_slices(self._slice_obj, self._fixup_slice(index))
        else:
            index = combine_slice_index(self._slice_obj, index)
        self.full_sequence[index] = value        
