import re
import os
import itertools
import weakref
from DataTypes.Bunch import Bunch
from DataTypes import CDataTypes
from DataTypes.ADict import ADict

class Error(Exception): pass
class MissingTypeError(Error): pass

def readelf_int(x):
    if x.startswith('0x'):
        return int(x, 16)
    else:
        return int(x)

def _symbols(exec_name):
    assert os.path.exists(exec_name)
    stdin, stdout = os.popen2('readelf --wide --symbols %s' % (exec_name,))
    for line in itertools.islice(stdout, 3, None):
        elements = line.split()
        if len(elements) == 8:
            num, off, size, type, bind, vis, ndx, name = elements
            # sometimes the table header line is re-printed, so we have to ignore it
            if num != "Num:":
                off = int(off, 16)
                size = readelf_int(size)
                yield name, (off, size, type)

def symbols(exec_name):
    return dict(_symbols(exec_name))

debug_dump_re = \
r""".*?The section \.debug_info contains:

(.*?)


.*"""
debug_dump_pattern = re.compile(debug_dump_re, re.DOTALL)

debug_info_re = \
r"""  Compilation Unit @ (?:offset 0x)?([\da-fA-F]+):
(.*?)
?(?=  Compilation Unit|$)"""
debug_info_pattern = re.compile(debug_info_re, re.DOTALL)

compilation_unit_re = \
r"""   Length:        (\d+)
   Version:       (\d+)
   Abbrev Offset: (\d+)
   Pointer Size:  (\d+)
(.*)"""
compilation_unit_pattern = re.compile(compilation_unit_re, re.DOTALL)

tag_re = \
r""" <(\d+)><([\da-fA-F]+)>: Abbrev Number: (\d+) \(DW_TAG_([a-zA-Z_]+?)\)
(.*?)(?= <\d+><[\da-fA-F]+>|$)"""
tag_pattern = re.compile(tag_re, re.DOTALL)

attr_re = \
r"""\s*DW_AT_([a-zA-Z_]+)\s*:\s*(?:\(indirect string, offset: 0x(?:[\da-fA-F]+)\): )?(.*?)\s*(?=\n|$)"""
attr_pattern = re.compile(attr_re, re.DOTALL)

tag_ref_re = \
r"""<([\da-fA-F]+)>"""
tag_ref_pattern = re.compile(tag_ref_re, re.DOTALL)

class TagMap(dict):
    pass

class CompilationUnit(Bunch): pass
class Tag(Bunch): pass

class Variable(Bunch): pass

class TagRef(object):
    def __init__(self, tagmap, offset):
        self._tagmap = weakref.proxy(tagmap)
        self.offset = offset
    def tag(self):
        return self._tagmap[self.offset]
    def __repr__(self):
        return '<TagRef 0x%X>' % (self.offset,)

base_type_of_name = {
    'char' : CDataTypes.Char,
    'signed char' : CDataTypes.Char,
    'unsigned char' : CDataTypes.UnsignedChar,
    'unsigned int' : CDataTypes.UnsignedInt,
    'int' : CDataTypes.Int,
    'float' : CDataTypes.Float,
    'double' : CDataTypes.Double,
    'short int' : CDataTypes.Short,
    'short unsigned int' : CDataTypes.UnsignedShort,
    'long int' : CDataTypes.Long,
    'long unsigned int' : CDataTypes.UnsignedLong,
    'long long int' : CDataTypes.LongLong,
    'long long unsigned int' : CDataTypes.UnsignedLongLong,
}

def _guess_packing_level(field_container, byte_size):
    for attempted_pack_level in [1, 2, 4]:
        field_container.packing_level = attempted_pack_level
        if field_container.size() == byte_size:
            break
    else:
        raise ValueError("Cannot find packing level for %r to be of size %r" % (field_container.field_types, byte_size))

def get_int(mappable, key, *args, **kw):
    x = mappable.get(key)
    if x is None:
        return None
    return readelf_int(x, *args, **kw)

type_info_names = [
    'variables',
    'typedefs',
    'structs',
    'unions',
    'enums',
]

def create_type_dicts(x):
    for type_info_name in type_info_names:
        setattr(x, type_info_name, ADict())

class ElfInfo(object):
    def __init__(self, exec_name):
        assert os.path.exists(exec_name)
        self.exec_name = exec_name

        self.symbols = ADict(symbols(self.exec_name))
        self.name_of_symbol_offset = dict(
            (offset, name)
            for name, (offset, size, type) in self.symbols.iteritems())

        stdin, stdout = os.popen2('readelf --wide --debug-dump %s' % (exec_name,))
        data = stdout.read()

        self.tagmap = TagMap()
        match = debug_dump_pattern.match(data)
        if match is not None:
            (self.compilation_units_str,) = match.groups()
            self._parse_compilation_units()

            self._build_high_level_description()
    
    def symbol_area(self, name):
        if name not in self.symbols:
            return None
        offset, size, type = self.symbols[name]
        return offset, size

    def _build_type__skip_into(self, tag):
        type_ = tag.attrs.get('type')
        if type_ is not None:
            self.type_of_tag_offset[tag.offset] = self._build_type(type_.tag())

    _build_type__const_type = _build_type__skip_into
    _build_type__volatile_type = _build_type__skip_into

    def _build_type__base_type(self, tag):
        type_ = base_type_of_name.get(tag.attrs['name'])
        if type_ is None:
            return
        assert type_.size() == readelf_int(tag.attrs['byte_size'])
        self.type_of_tag_offset[tag.offset] = type_

    def _build_type__field_container(self, base_type, container, tag):
        # TODO: This is becoming a _huge_ function, divide!

        byte_size = get_int(tag.attrs, 'byte_size')

        # Verify no bitfields, because we don't support it yet...
        for child in tag.children:
            if 'bit_size' in child.attrs:
                # Its a bitfield member
                # We don't currently support bitfields in structs.. :-(
                # Try to use a fake buffer of the same size for now..
                if byte_size is not None:
                    type_ = CDataTypes.buffer_size(byte_size)
                    self.type_of_tag_offset[tag.offset] = type_
                return

        name = tag.attrs.get('name', 'Anonymous%s_%X' % (base_type.__name__, tag.offset))

        if name not in container:
            class type_(base_type):
                # Until we know the field_types
                field_types = None
            self.type_of_tag_offset[tag.offset] = type_
            type_.__name__ = name

            # Add the struct/union to the big struct/union dict
            container[name] = type_
        type_ = container[name]
        
        if tag.attrs.get('declaration') == '1':
            assert not tag.children, "forward declared %r had values %r" % (type_, tag)
        else:
            assert type_.field_types is None, "%r redeclared by %r" % (type_, tag)
            def field_type(child):
                if child.name != 'member':
                    self._build_type(child)
                    return None
                name = child.attrs['name']
                type_ = self._get_type(child.attrs['type'].tag())
                return (name, type_)
            type_.field_types = filter(None, [field_type(child) for child in tag.children])

            _guess_packing_level(type_, byte_size)

    def _build_type__union_type(self, tag):
        container = self.current_compilation_unit.unions
        self._build_type__field_container(CDataTypes.Union, container, tag)

    def _build_type__structure_type(self, tag):
        container = self.current_compilation_unit.structs
        self._build_type__field_container(CDataTypes.Struct, container, tag)

    def _build_type__array_type(self, tag):
        node_type = self._build_type(tag.attrs['type'].tag())
        if node_type is None:
            return
        while tag.children:
            subrange_tag = tag.children.pop(-1)
            upper_bound_str = subrange_tag.attrs.get('upper_bound', None)

            # This means the array size is a variable (i.e C99's int x[y])
            # and we don't support it!
            c99_var_array = isinstance(upper_bound_str, TagRef)
            if c99_var_array:
                array_count = None
            elif upper_bound_str is None:
                # gcc seems to put no size field at all, when the array size is
                # [0] lets assume that this works backwards, and that if no
                # size is given [0] is meant.
                array_count = 0
            else:
                # Sometimes there is crap after the number. readelf is a piece of
                # shit.  Get rid of the shit and take only the number.
                upper_bound_num_str = upper_bound_str.split(None, 1)[0]
                upper_bound = readelf_int(upper_bound_num_str)
                array_count = upper_bound + 1
            class Array(CDataTypes.Array):
                count = array_count
            Array.node_type = node_type
            Array.__name__ = 'Array_%s_%s' % (array_count, Array.node_type.__name__)
            node_type = Array
        self.type_of_tag_offset[tag.offset] = node_type

    def _build_type__typedef(self, tag):
        type_tag = tag.attrs.get('type')
        if type_tag is not None:
            type_ = self._build_type(type_tag.tag())
            if type_ is None:
                return
            name = tag.attrs['name']
            typedef = CDataTypes.typedef(name, type_)
            self.type_of_tag_offset[tag.offset] = typedef
            self.current_compilation_unit.typedefs[name] = typedef
        
    def _build_type__pointer_type(self, tag):
        class Pointer(CDataTypes.Pointer):
            pass
        Pointer.__name__ = 'PointerToUnknown'
        self.type_of_tag_offset[tag.offset] = Pointer
        type_tag = tag.attrs.get('type')
        if type_tag is not None:
            def fix_pointer():
                type_ = self._build_type(type_tag.tag())
                if type_ is None:
                    return
                Pointer.to_type = type_
                Pointer.__name__ = 'PointerTo%s' % (type_.__name__,)
            self.pointers_fixup.append(fix_pointer)

    def _build_type__enumeration_type(self, tag):
        name = tag.attrs.get('name', 'AnonymousEnum_%X' % (tag.offset,))
        container = self.current_compilation_unit.enums
        if name not in container:
            value_names = None
            enum = CDataTypes.enum_of(CDataTypes.Int, value_names, name)
            container[name] = enum
        enum = container[name]
        if tag.attrs.get('declaration') == '1':
            assert not tag.children, "forward declared %r had values %r" % (enum, tag)
        else:
            assert enum.value_names is None, "%r redeclared by %r" % (enum, tag)
            def get_const_value(child):
                # GCC has a bug of not putting the "const_value" field of
                # negative enum constants.  We put None when these are missing,
                # so we can crash later, only if we actually need the value.
                const_value = child.attrs.get('const_value', None)
                if const_value is not None:
                    const_value = readelf_int(const_value)
                return (child.attrs['name'], const_value)
            value_names = [get_const_value(child) for child in tag.children]
            enum.set_value_names(value_names)
        self.type_of_tag_offset[tag.offset] = enum

    def _build_type__subroutine_type(self, tag):
        # TODO: Build a function pointer, use ABI and param information!
        pass


    def _build_type(self, tag):
        if tag.offset not in self.type_of_tag_offset:
            builder = getattr(self, '_build_type__%s' % (tag.name,), None)
            assert builder is not None, "No builder for %r at tag %X" % (tag.name, tag.offset)
            builder(tag)
        return self.type_of_tag_offset.get(tag.offset, None)

    def _get_type(self, tag):
        type_ = self._build_type(tag)
        if type_ is None:
            raise MissingTypeError("Missing type for tag", tag)
        return type_
    
    def _build_high_level_description(self):
        self.type_of_tag_offset = ADict()
        create_type_dicts(self)
        # We don't want to follow pointers so we can avoid finding structs we
        # are currently building via pointers
        # Thus we fix the pointers later...
        for compilation_unit in self.compilation_units:
            self._build_high_level_description_compilation_unit(compilation_unit)
            for type_info_name in type_info_names:
                compilation_unit_type_info = getattr(compilation_unit, type_info_name)
                getattr(self, type_info_name).update(compilation_unit_type_info)

    def _build_high_level_description_compilation_unit(self, compilation_unit):
        self.current_compilation_unit = compilation_unit
        self.pointers_fixup = []
        create_type_dicts(self.current_compilation_unit)
        self._build_high_level_description_tags(compilation_unit.tags)
        while self.pointers_fixup:
            fixup = self.pointers_fixup.pop()
            fixup()
        del self.current_compilation_unit

    def _build_high_level_description_tags(self, tags):
        for tag in tags:
            if tag.name == 'variable' and tag.attrs.get('artificial', 0) == 0:
                name = tag.attrs.get('name')
                if name is not None:
                    variable = Variable(
                        name = name,
                        decl_place = (tag.attrs['decl_file'],
                                      tag.attrs['decl_line']),
                        type = self._build_type(tag.attrs['type'].tag())
                    )
                    self.current_compilation_unit.variables[name] = variable
            elif tag.name == 'typedef' and 'type' in tag.attrs:
                self._build_type(tag)
            elif tag.name in ['enumeration_type', 'struct_type', 'union_type']:
                self._build_type(tag)
            self._build_high_level_description_tags(tag.children)

    def _parse_compilation_units(self):
        self.compilation_units = [
            self._compilation_unit(int(match.group(1), 16), match.group(2))
            for match in debug_info_pattern.finditer(self.compilation_units_str)
        ]
    
    def tag_value(self, value):
        match = tag_ref_pattern.match(value)
        if match:
            return TagRef(self.tagmap, int(match.group(1), 16))
        return value

    def _tag(self, offset, abbrev_num, name, text):
        attrs=dict((match.group(1), self.tag_value(match.group(2)))
                   for match in attr_pattern.finditer(text))
        if 'sibling' in attrs:
            # This attribute already manifests itself in the tree
            # structure.
            del attrs['sibling']
        tag = Tag(offset=offset,
                  abbrev_num=abbrev_num,
                  name=name,
                  attrs=attrs)
        self.tagmap[offset] = tag
        return tag

    def _compilation_unit(self, offset, compilation_unit_text):
        match = compilation_unit_pattern.match(compilation_unit_text)
        assert match, text
        length, version, abbrev_offset, ptr_size, tags_text = match.groups()
        length, version, abbrev_offset, ptr_size = [
            int(x) for x in
            [length, version, abbrev_offset, ptr_size]
        ]
        compilation_unit = CompilationUnit(
            offset=offset,
            length=length,
            version=version,
            abbrev_offset=abbrev_offset,
            ptr_size=ptr_size,
            tags=[],
        )
        parents_of_level = [compilation_unit.tags]
        for match in tag_pattern.finditer(tags_text):
            indent_level, offset, abbrev_num, name, text = (
                int(match.group(1)), int(match.group(2), 16),
                int(match.group(3)), match.group(4),
                match.group(5))
            del parents_of_level[indent_level+1:]
            parent = parents_of_level[indent_level]
            tag = self._tag(offset, abbrev_num, name, text)
            tag.children = []
            parents_of_level.append(tag.children)
            parent.append(tag)
        compilation_unit.attrs = compilation_unit.tags[0].attrs
        return compilation_unit
